import csv
import json
import logging
import os
from typing import List

import numpy as np
import psycopg2
from qdrant_client import QdrantClient
from transformers import AutoModel

from agent_search.core import AgentSearchResult
from agent_search.core.utils import (
    cosine_similarity,
    get_data_path,
    load_config,
)

logger = logging.getLogger(__name__)


class WebSearchEngine:
    """A simple search client for the OpenSearch collection"""

    def __init__(
        self,
    ):
        try:
            import psycopg2
        except ImportError as e:
            raise ImportError(
                f"Error {e} while imoprting psycopg2. Please install it with `pip install psycopg2` to run an WebSearchEngine instance."
            )

        # Load config
        self.config = load_config()["agent_search"]

        # Load Postgres
        logger.info(
            f"Connecting to Postgres database at: {self.config['postgres_db']}."
        )

        # Load qdrant client
        logger.info(
            f"Connecting to collection: {self.config['qdrant_collection_name']}"
        )
        self.qdrant_collection_name = self.config["qdrant_collection_name"]
        self.client = QdrantClient(
            self.config["qdrant_host"],
            grpc_port=self.config["qdrant_grpc_port"],
            prefer_grpc=True,
        )
        if not self.client.get_collection(self.qdrant_collection_name):
            raise ValueError(
                f"Must have a Qdrant collection with the name {self.qdrant_collection_name}."
            )

        # Load embedding model
        self.embedding_model = AutoModel.from_pretrained(
            self.config["embedding_model_name"], trust_remote_code=True
        )

        self.pagerank_rerank_module = self.config["pagerank_rerank_module"]
        pagerank_file_path = self.config["pagerank_file_path"]
        if self.pagerank_rerank_module:
            if not pagerank_file_path:
                # Simulating reading from a CSV file
                pagerank_file_path = os.path.join(
                    get_data_path(), "domain_ranks.csv"
                )

                if not os.path.exists(pagerank_file_path):
                    raise ValueError(
                        "Must have a pagerank file at the config specified path when using pagerank_rerank_module"
                    )

            self.pagerank_importance = float(
                self.config["pagerank_importance"]
            )
            self.domain_to_rank_map = {}

            with open(pagerank_file_path, newline="") as csvfile:
                reader = csv.DictReader(csvfile)
                for row in reader:
                    domain = row["Domain"]
                    rank = float(row["Open Page Rank"])
                    self.domain_to_rank_map[domain] = rank

    def get_query_vector(self, query: str):
        """Gets the query vector for the given query"""

        query_vector = self.embedding_model.encode(query)
        return query_vector

    def similarity_search(
        self,
        query_vector: np.ndarray,
        limit: int = 100,
    ):
        """Searches the collection for the given query and returns the top 'limit' results"""

        points = self.client.search(
            collection_name=self.qdrant_collection_name,
            query_vector=query_vector,
            limit=limit,
        )

        results = []
        for point in points:
            try:
                results.append(
                    AgentSearchResult(
                        score=point.score,
                        text=point.payload["text"],
                        title=None,
                        url=point.payload["url"],
                        metadata={},
                    )
                )
            except Exception as e:
                logger.error(f"Error appending point {point} with {e}")
        return results

    # Example of batch processing
    def execute_batch_query(self, urls, batch_size=20):
        results = []
        try:
            with psycopg2.connect(
                dbname=self.config["postgres_db"],
                user=self.config["postgres_user"],
                password=self.config["postgres_password"],
                host=self.config["postgres_host"],
                options="-c client_encoding=UTF8",
            ) as conn:
                with conn.cursor() as cur:
                    for i in range(0, len(urls), batch_size):
                        batch_urls = urls[i : i + batch_size]
                        logger.info(
                            f"Executing batch query for URLs: {batch_urls[0:2]}"
                        )
                        query = f"SELECT url, title, metadata, dataset, text_chunks, embeddings FROM {self.config['postgres_table_name']} WHERE url = ANY(%s)"
                        cur.execute(query, (batch_urls,))
                        batch_results = cur.fetchall()
                        results.extend(batch_results)
        except psycopg2.DatabaseError as e:
            logger.error(f"Database error: {e}")
        except Exception as e:
            logger.error(f"Error in execute_batch_query: {e}")
        return results

    def hierarchical_similarity_reranking(
        self,
        query_vector: np.ndarray,
        urls: List[str],
        limit: int = 100,
    ) -> List[AgentSearchResult]:
        """Hierarchical URL search to find the most similar text chunk for the given query and URLs"""
        results = self.execute_batch_query(urls)
        # List to store the results along with their similarity scores
        similarity_results = []

        # Iterate over each result to find the most similar text chunk
        for result in results:
            (
                url,
                title,
                metadata,
                dataset,
                text_chunks_str,
                embeddings_binary,
            ) = result
            # deserialize the embeddings and text chunks
            embeddings = np.frombuffer(
                embeddings_binary, dtype=np.float32
            ).reshape(-1, 768)
            text_chunks = json.loads(text_chunks_str)
            max_similarity = -1e9
            most_similar_chunk = None

            # Iterate over each embedding to find the one with maximum cosine similarity
            for chunk, embedding in zip(text_chunks, embeddings):
                similarity = cosine_similarity(
                    np.array(query_vector), np.array(embedding)
                )
                if similarity > max_similarity:
                    max_similarity = similarity
                    most_similar_chunk = chunk

            # Store the most similar chunk and its similarity score
            similarity_results.append(
                AgentSearchResult(
                    score=max_similarity,
                    url=url,
                    title=title,
                    metadata=json.loads(metadata),
                    dataset=dataset,
                    text=most_similar_chunk,
                ),
            )

        # Sort the results based on similarity score in descending order
        similarity_results.sort(key=lambda x: x.score, reverse=True)
        return similarity_results[:limit]

    def pagerank_reranking(
        self,
        similarity_results: List[AgentSearchResult],
        limit: int = 100,
    ) -> List[AgentSearchResult]:
        """Reranks the results based on the PageRank score of the domain"""
        if not self.pagerank_rerank_module:
            raise Exception(
                "PageRank reranking module is not enabled. Please set pagerank_rerank_module=True while initializing the WebSearchEngine client."
            )
        # List to store the results along with their PageRank scores
        pagerank_results = []

        # Iterate over each result to find the PageRank score of the domain
        for result in similarity_results:
            pagerank_score = 0
            try:
                domain = result.url.split("/")[2]
                pagerank_score = self.domain_to_rank_map.get(domain, 0)
            except Exception as e:
                logger.info(f"Error {e}: Found for URL: {result.url}")
            reweighted_score = (
                self.pagerank_importance * pagerank_score / 10.0
                + (1 - self.pagerank_importance) * result.score
            )
            pagerank_results.append(
                AgentSearchResult(
                    score=reweighted_score,
                    url=result.url,
                    title=result.title,
                    metadata=result.metadata,
                    dataset=result.dataset,
                    text=result.text,
                )
            )

        # Sort the results based on PageRank score in descending order
        pagerank_results.sort(key=lambda x: x.score, reverse=True)
        return pagerank_results[:limit]



import configparser
import os
from typing import List, Optional

import numpy as np

from agent_search.core import AgentSearchResult


def select_top_urls(
    ordered_points: List[AgentSearchResult],
    max_urls: int = 10,
    url_contains: Optional[List[str]] = None,
) -> List[str]:
    """A function to return the top unique URLs from the given poitns results."""
    if not url_contains:
        url_contains = []

    top_urls = set([])
    for point in ordered_points:
        url = point.url
        if url in top_urls:
            continue
        url_contains_match = False if url_contains else True
        for url_contain in url_contains:
            if url_contain in url:
                url_contains_match = True
                break
        if not url_contains_match:
            continue
        top_urls.add(point.url)
        if len(top_urls) >= max_urls:
            break

    return list(top_urls)


def cosine_similarity(v1: np.ndarray, v2: np.ndarray) -> float:
    """Compute the cosine similarity between two vectors."""
    dot_product = np.dot(v1, v2)
    norm_v1 = np.linalg.norm(v1)
    norm_v2 = np.linalg.norm(v2)
    return dot_product / (norm_v1 * norm_v2)


def get_data_path() -> str:
    return os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "data",
    )


def load_config(config_dir: Optional[str] = None) -> configparser.ConfigParser:
    """Load the configuration file."""
    config = configparser.ConfigParser()
    if not config_dir:
        config_dir = get_data_path()
    config.read(os.path.join(config_dir, "config.ini"))
    return config



import json
import logging
import os
import time
from typing import Any, Dict, List, Optional

import httpx
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)


class SearchResult(BaseModel):
    score: Optional[float] = None
    url: str
    title: str
    text: str
    dataset: str
    metadata: Any


class SearchRAGResponse(BaseModel):
    response: str
    related_queries: List[str]
    search_results: List[SearchResult]


class SciPhi:
    """
    Client for interacting with the SciPhi API.

    Attributes:
        api_base (str): Base URL for the SciPhi API.
        api_key (str): API key for authenticating requests.
        timeout (int): Timeout for API requests in seconds.
        client (httpx.Client): HTTP client for making requests.
    """

    def __init__(
        self,
        api_base: Optional[str] = None,
        api_key: Optional[str] = None,
        timeout: int = 30,
    ) -> None:
        """
        Initializes the SciPhi client.

        Args:
            api_base (Optional[str]): Base URL for the SciPhi API.
            api_key (Optional[str]): API key for authenticating requests.
            timeout (int): Timeout for API requests in seconds.

        Raises:
            ValueError: If `api_key` is not provided.
        """

        self.api_base = (
            api_base or os.getenv("SCIPHI_API_BASE") or "https://api.sciphi.ai"
        )
        self.api_key = api_key or os.getenv("SCIPHI_API_KEY")
        if not self.api_key:
            raise ValueError(
                "You must specify an explicit api_key or define `SCIPHI_API_KEY` to initialize a SciPhi client."
            )
        self.timeout = timeout
        self.client = httpx.Client(
            base_url=self.api_base,
            headers=self._auth_headers(),
            timeout=timeout,
        )

    def _auth_headers(self) -> Dict[str, str]:
        """
        Generates the authorization headers for the API requests.

        Returns:
            Dict[str, str]: Authorization headers with bearer token.
        """

        return {"Authorization": f"Bearer {self.api_key}"}

    def _handle_api_response(self, response: httpx.Response) -> Dict:
        """
        Handles the HTTP response from the API.

        Args:
            response (httpx.Response): The response from the API request.

        Returns:
            Dict: JSON response content.

        Raises:
            Exception: If the response indicates an error.
        """

        if response.is_error:
            # Handle errors appropriately
            raise Exception(
                f"API request failed with status {response.status_code}"
            )
        result = response.json()
        return result

    def _handle_search_response(self, search_results: Dict[str, str]) -> None:
        """
        Handles dictionary search resopnses from the API.

        Args:
            search_results (Dict[str, str]): The response from the API request.

        Returns:
            Dict: JSON response content.

        Raises:
            Exception: If the response indicates an error.
        """

        for result in search_results:
            if "score" in result:
                result["score"] = float(result["score"])
            if "metadata" in result:
                try:
                    result["metadata"] = (
                        json.loads(result["metadata"])
                        if (
                            result["metadata"] != None
                            and result["metadata"] != '""'
                        )
                        else {}
                    )
                except Exception as e:
                    result["metadata"] = dict()

    def _retry_api_request(
        self, method: str, url: str, payload: Dict, max_retries: int = 3
    ):
        """
        Common method for retrying API requests with exponential backoff.

        Args:
            method (str): The HTTP method to use ('get' or 'post').
            url (str): The API endpoint.
            payload (Dict): The payload for the request.
            max_retries (int): Maximum number of retry attempts.

        Returns:
            Dict: The JSON response from the API.

        Raises:
            Exception: If the maximum number of retries is reached.
        """
        for attempt in range(max_retries):
            try:
                response = getattr(self.client, method)(url, json=payload)
                return self._handle_api_response(response)

            except httpx.HTTPError as e:
                logger.info(f"HTTP error on attempt {attempt + 1}: {e}")
                if attempt < max_retries - 1:
                    time.sleep(0.5 * (2**attempt))

            except Exception as e:
                logger.error(f"Error on attempt {attempt + 1}: {e}")
                if attempt < max_retries - 1:
                    time.sleep(0.5 * (2**attempt))

        raise Exception("Failed to fetch data after maximum retries.")

    def search(
        self, query: str, search_provider: str, max_retries: int = 3
    ) -> List[Dict]:
        """
        Performs a search query using the SciPhi API with retry and backoff logic.

        Args:
            query (str): The search query string.
            search_provider (str): The search provider to use.
            max_retries (int): Maximum number of retry attempts.

        Returns:
        List[Dict]: A list of search results.
        """
        url = f"/search"
        payload = {"provider": search_provider, "query": query}
        try:
            handled_response = self._retry_api_request(
                "post", url, payload, max_retries
            )
            self._handle_search_response(handled_response)
            return [SearchResult(**ele).dict() for ele in handled_response]

        except Exception as e:
            logger.error(f"Search request failed: {e}")
            return {"error": str(e)}

    def get_search_rag_response(
        self,
        query: str,
        search_provider: str,
        llm_model: str = "SciPhi/Sensei-7B-V1",
        temperature: int = 0.2,
        top_p: int = 0.95,
    ):
        """
        Retrieves a search RAG (Retrieval-Augmented Generation) response from the API.

        Args:
            query (str): The search query string.
            search_provider (str): The search provider to use.
            llm_model (str): The language model to use.
            temperature (int): The temperature setting for the query.
            top_p (int): The top-p setting for the query.

        Returns:
            Dict: A dictionary with the search response and related queries.
        """

        if query == "":
            raise ValueError("Blank query submitted.")
        if search_provider not in ["bing", "agent-search"]:
            raise ValueError(f"Unsupported provider, {search_provider}")

        url = f"/search_rag"
        payload = {
            "query": query,
            "search_provider": search_provider,
            "llm_model": llm_model,
            "temperature": temperature,
            "top_p": top_p,
        }
        try:
            handled_response = self._retry_api_request("post", url, payload)

            # rename the other queries to `related_queries` until LLM output is re-factored.
            handled_response["related_queries"] = handled_response.pop(
                "other_queries"
            )

            self._handle_search_response(handled_response["search_results"])
            # Use Pydantic model for parsing and validation
            search_response = SearchRAGResponse(**handled_response)
        except Exception as e:
            logger.error(f"Search request failed: {e}")
            return {"error": str(e)}

        return search_response.dict()

    def completion(
        self,
        prompt: str,
        llm_model_name: str = "SciPhi/Sensei-7B-V1",
        llm_max_tokens_to_sample: int = 1_024,
        llm_temperature: float = 0.2,
        llm_top_p: float = 0.90,
    ) -> SearchRAGResponse:
        """
        Generates a completion for a given prompt using the SciPhi API.

        Args:
            prompt (str): The prompt for generating completion.
            llm_model_name (str): The language model to use.
            llm_max_tokens_to_sample (int): Maximum number of tokens for the sample.
            llm_temperature (float): The temperature setting for the query.
            llm_top_p (float): The top-p setting for the query.

        Returns:
            Dict: A dictionary containing the generated completion.

        Raises:
            ImportError: If the `sciphi-synthesizer` package is not installed.
        """

        try:
            import synthesizer
        except ImportError as e:
            raise ImportError(
                "Please install run `pip install sciphi-synthesizer` before attempting to generate a completion."
            )

        from synthesizer.core import LLMProviderName
        from synthesizer.interface import LLMInterfaceManager
        from synthesizer.llm import GenerationConfig

        try:
            llm_interface = LLMInterfaceManager.get_interface_from_args(
                LLMProviderName("sciphi"),
            )

            generation_config = GenerationConfig(
                model_name=llm_model_name,
                max_tokens_to_sample=llm_max_tokens_to_sample,
                temperature=llm_temperature,
                top_p=llm_top_p,
            )

            completion = llm_interface.get_completion(
                prompt, generation_config
            ).replace("</s>", "")

            return completion

        except Exception as e:
            logger.error(f"Completion generation failed: {e}")
            return {"error": str(e)}

    def close(self) -> None:
        """
        Closes the HTTP client.
        """

        self.client.close()



from typing import Any, Optional

from pydantic import BaseModel


class AgentSearchResult(BaseModel):
    """A dataclass to store the search result"""

    score: float
    url: str
    title: Optional[str]
    dataset: Optional[str]
    # TODO - Add dict(str, [str, float, ..]) validation
    metadata: Any
    text: str

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.title and self.title == self.text[0 : len(self.title)]:
            self.text = self.text[len(self.title) :]
        self.text = self.text.strip()

    def to_string_dict(self) -> dict:
        """Returns a dictionary representation with all values as strings."""
        return {
            "score": str(self.score),
            "url": self.url,
            "title": self.title,
            "dataset": self.dataset,
            "metadata": self.metadata,
            "text": self.text,
        }

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)



"""A script to populate the database with the given dataset and subset."""
import json
import logging
import multiprocessing
import uuid

import fire
import numpy as np
import psycopg2
from qdrant_client import QdrantClient
from qdrant_client.http import models

from agent_search.core.utils import load_config

logger = logging.getLogger(__name__)

EMBEDDING_VEC_SIZE = 768


def create_collection(qclient, collection_name):
    qclient.create_collection(
        collection_name=collection_name,
        vectors_config=models.VectorParams(
            size=768, distance=models.Distance.COSINE
        ),
        quantization_config=models.ScalarQuantization(
            scalar=models.ScalarQuantizationConfig(
                type=models.ScalarType.INT8,
                quantile=0.99,
                always_ram=True,
            ),
        ),
    )


def process_rows(rows, output_queue):
    """Process the rows into qdrant point objects."""
    qdrant_points = []
    for row in rows:
        _, url, __, text_chunks, embeddings_binary, ___, ____ = row
        embeddings = np.frombuffer(
            embeddings_binary, dtype=np.float32
        ).reshape(-1, EMBEDDING_VEC_SIZE)

        text_chunks = json.loads(text_chunks)
        # Prepare data for Qdrant
        qdrant_points.append(
            models.PointStruct(
                id=str(uuid.uuid3(uuid.NAMESPACE_DNS, url)),
                vector=[float(ele) for ele in embeddings[0]],
                payload={"text": text_chunks[0], "url": url},
            )
        )

    output_queue.put(qdrant_points)


def qdrant_writer(config, qdrant_queue, delete_existing):
    """A writer that listens for output events in a separate thread."""
    qclient = QdrantClient(
        config["qdrant_host"],
        port=config["qdrant_grpc_port"],
        prefer_grpc=config["qdrant_prefer_grpc"],
    )
    if delete_existing:
        qclient.delete_collection(config["qdrant_collection_name"])
        create_collection(qclient, config["qdrant_collection_name"])

    logger.info("Launching Qdrant writer")
    while True:
        try:
            points = qdrant_queue.get()
            logger.info(f"Starting Qdrant write-out...")
            if points is None:  # Sentinel to end the process
                break
            operation_result = qclient.upsert(
                collection_name=config["qdrant_collection_name"],
                wait=True,
                points=points,
            )
            logger.info(
                f"Finished Qdrant write-out with result {operation_result}..."
            )
        except Exception as e:
            logger.info(f"Task failed with {e}")


def process_batches(config, start, end, batch_size, output_queue):
    """Processes the batches in steps of the given batch_size"""

    # Connect to the database
    conn = psycopg2.connect(
        dbname=config["postgres_db"],
        user=config["postgres_user"],
        password=config["postgres_password"],
        host=config["postgres_host"],
        options="-c client_encoding=UTF8",
    )
    cur = conn.cursor()
    # Declare a server-side cursor with offset
    cur.execute(
        f"DECLARE proc_cursor CURSOR FOR SELECT * FROM {config['postgres_table_name']} OFFSET {start} LIMIT {end - start}"
    )

    offset = start
    while True:
        logger.info(
            f"Fetching a batch of size {batch_size} at offset {offset}"
        )
        # Fetch a batch of rows
        cur.execute(f"FETCH {batch_size} FROM proc_cursor")
        rows = cur.fetchall()

        if len(rows) == 0:
            break

        process_rows(rows, output_queue)
        offset += batch_size

        # terminate
        if offset + batch_size >= end:
            break

    cur.close()
    conn.close()


class PopulateQdrant:
    def __init__(self):
        self.config = load_config()["agent_search"]

    def run(self, num_processes=16, batch_size=1_024, delete_existing=False):
        """Runs the population process for the qdrant database"""
        qdrant_queue = multiprocessing.Queue()
        qdrant_writer_thread = multiprocessing.Process(
            target=qdrant_writer,
            args=(
                self.config,
                qdrant_queue,
                delete_existing,
            ),
        )
        qdrant_writer_thread.start()

        conn = psycopg2.connect(
            dbname=self.config["postgres_db"],
            user=self.config["postgres_user"],
            password=self.config["postgres_password"],
            host=self.config["postgres_host"],
            options="-c client_encoding=UTF8",
        )
        cur = conn.cursor()

        # Count total number of entries
        cur.execute(
            f"SELECT COUNT(*) FROM {self.config['postgres_table_name']}"
        )
        total_count = cur.fetchone()[0]
        logger.info(
            f"Processing {total_count} entries in {num_processes} processes"
        )

        range_size = total_count // num_processes

        # Create and start multiprocessing workflow
        processes = []
        for i in range(num_processes):
            logger.info(f"Starting process {i}...")
            start = i * range_size
            end = start + range_size if i < num_processes - 1 else total_count
            proc = multiprocessing.Process(
                target=process_batches,
                args=(
                    self.config,
                    start,
                    end,
                    batch_size,
                    qdrant_queue,
                ),
            )
            processes.append(proc)
            proc.start()

        # Wait for all processes to finish
        for proc in processes:
            proc.join()

        # send termination signal
        qdrant_queue.put(None)

        cur.close()
        conn.close()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logger.setLevel(logging.INFO)
    fire.Fire(PopulateQdrant)



import os
from typing import List, Optional

import requests

from .search_types import AgentSearchResult


class AgentSearchClient:
    def __init__(
        self,
        api_base: Optional[str] = None,
        auth_token: Optional[str] = None,
    ):
        self.api_base = (
            api_base or os.getenv("SCIPHI_API_BASE") or "https://api.sciphi.ai"
        )
        self.auth_token = auth_token or os.getenv("SCIPHI_API_KEY")

        if not self.auth_token:
            raise ValueError(
                "No authorization token provided and SCIPHI_API_KEY environment variable is not set."
            )

    def search(
        self,
        query: str,
        limit_broad_results: int = 1_000,
        limit_deduped_url_results: int = 100,
        limit_hierarchical_url_results: int = 25,
        limit_final_pagerank_results: int = 10,
    ) -> List[AgentSearchResult]:
        headers = {
            "Authorization": f"Bearer {self.auth_token}",
            "Content-Type": "application/json",
        }
        payload = {
            "query": query,
            "limit_broad_results": limit_broad_results,
            "limit_deduped_url_results": limit_deduped_url_results,
            "limit_hierarchical_url_results": limit_hierarchical_url_results,
            "limit_final_pagerank_results": limit_final_pagerank_results,
        }
        response = requests.post(
            f"{self.api_base}/search", headers=headers, json=payload
        )
        response.raise_for_status()  # Raises an HTTPError if the HTTP request returned an unsuccessful status code
        results = response.json()

        serp_results = [
            AgentSearchResult.from_dict(result) for result in results
        ]

        return serp_results



import json

import fire

PROMPT = """
### Instruction:

Query:
{query}

Search Results:
{rag_context}

Query:
{query}

### Response:
{{"response":
"""

PROMPT = """
### Instruction:

Query:
{query}

Search Results:
{rag_context}

Query:
{query}

### Response:
{{"response":
"""


class RagDemo:
    """A demonstration of Bing + synthesizer RAG pipeline."""

    def __init__(self):
        try:
            import synthesizer
        except ImportError as e:
            raise ImportError(
                f"Demo run_rag.py failed with {e}. Please run pip install sciphi-synthesizer before attempting to run this script."
            )

    def run(
        self,
        query="What is a quantum field theory in curved space time?",
        # Bing RAG provider settings
        rag_provider_name="bing",
        rag_api_base="https://api.bing.microsoft.com/v7.0/search",
        # llm parameters
        llm_provider_name="sciphi",
        llm_model_name="SciPhi/Sensei-7B-V1",
        llm_max_tokens_to_sample=1_024,
        llm_temperature=0.2,
        llm_top_p=0.90,
    ):
        from synthesizer.core import LLMProviderName, RAGProviderName
        from synthesizer.interface import (
            LLMInterfaceManager,
            RAGInterfaceManager,
        )
        from synthesizer.llm import GenerationConfig

        # Initialize Bing RAG Interface with its configuration
        rag_interface = RAGInterfaceManager.get_interface_from_args(
            provider_name=RAGProviderName(rag_provider_name),
            api_base=rag_api_base,
        )
        rag_result = rag_interface.get_rag_context(query)

        # LLM Provider Settings
        llm_interface = LLMInterfaceManager.get_interface_from_args(
            LLMProviderName(llm_provider_name),
        )

        generation_config = GenerationConfig(
            model_name=llm_model_name,
            max_tokens_to_sample=llm_max_tokens_to_sample,
            temperature=llm_temperature,
            top_p=llm_top_p,
            # other generation params here ...
        )

        formatted_prompt = PROMPT.format(
            rag_context=rag_result.context, query=query
        )
        completion = '{"response":' + llm_interface.get_completion(
            formatted_prompt, generation_config
        ).replace("</s>", "")

        print(
            f"Search Results:\n{rag_result.meta_data}"
            + f"\nPrompt:\n{formatted_prompt}\n\n"
            + "-" * 100
            + f"\nCompletion:\n{json.loads(completion)}"
        )


if __name__ == "__main__":
    fire.Fire(RagDemo)



import logging
import time
from typing import Optional

from pydantic import BaseModel

from agent_search.core.utils import load_config, select_top_urls
from agent_search.search import WebSearchEngine

# Attempt to import uvicorn and FastAPI
try:
    import uvicorn
    from fastapi import FastAPI, HTTPException
except ImportError as e:
    raise ImportError(
        f"Error: {e}, Note - both uvicorn and FastAPI are required to run the server."
    )


logger = logging.getLogger(__name__)


class SearchServer:
    def __init__(self):
        self.client = WebSearchEngine()

    def run(
        self,
        query="What is a lagrangian?",
        limit_broad_results=1_000,
        limit_deduped_url_results=50,
        limit_hierarchical_url_results=50,
        limit_final_pagerank_results=20,
        url_contains_filter=None,
    ):
        """Run a search query using the WebSearchEngine client"""

        query_vector = self.client.get_query_vector(query)

        broad_results = self.client.similarity_search(
            query_vector=query_vector, limit=limit_broad_results
        )

        if not url_contains_filter:
            url_contains_filter = []

        deduped_url_results = select_top_urls(
            broad_results,
            max_urls=limit_deduped_url_results,
            url_contains=url_contains_filter,
        )

        hierarchical_url_results = (
            self.client.hierarchical_similarity_reranking(
                query_vector=query_vector,
                urls=deduped_url_results,
                limit=limit_hierarchical_url_results,
            )
        )

        pagerank_reranked_results = self.client.pagerank_reranking(
            hierarchical_url_results
        )[:limit_final_pagerank_results]

        return pagerank_reranked_results


class SearchQuery(BaseModel):
    """A search query data model"""

    query: str
    limit_broad_results: Optional[int] = 1_000
    limit_deduped_url_results: Optional[int] = 100
    limit_hierarchical_url_results: Optional[int] = 25
    limit_final_pagerank_results: Optional[int] = 10


app = FastAPI()
search_runner = SearchServer()


def check_limits(query: SearchQuery):
    """Check if the limit parameters exceed three times their default values"""
    if query.limit_broad_results > 3 * 1_000:
        raise ValueError(
            "limit_broad_results exceeds 3 times its default value"
        )
    if query.limit_deduped_url_results > 3 * 100:
        raise ValueError(
            "limit_deduped_url_results exceeds 3 times its default value"
        )
    if query.limit_hierarchical_url_results > 3 * 25:
        raise ValueError(
            "limit_hierarchical_url_results exceeds 3 times its default value"
        )
    if query.limit_final_pagerank_results > 3 * 10:
        raise ValueError(
            "limit_final_pagerank_results exceeds 3 times its default value"
        )


@app.post("/search")
def run_search(query: SearchQuery):
    """Run a search query"""
    try:
        check_limits(query)
        results = search_runner.run(
            query=query.query,
            limit_broad_results=query.limit_broad_results,
            limit_deduped_url_results=query.limit_deduped_url_results,
            limit_hierarchical_url_results=query.limit_hierarchical_url_results,
            limit_final_pagerank_results=query.limit_final_pagerank_results,
        )
        return {"results": results}
    except ValueError as e:
        logger.error(f"ValueError {e} = ", e)
        raise HTTPException(status_code=400, detail=str(e))
    except Exception as e:
        logger.error(f"Exception {e} = ", e)
        raise HTTPException(status_code=500, detail=str(e))


@app.get("/health")
def health_check():
    """Health check endpoint"""
    return {"status": "ok"}


if __name__ == "__main__":
    config = load_config()["server"]
    logging.basicConfig(level=config["log_level"])
    uvicorn.run(app, host=config["host"], port=int(config["port"]))
